#!/usr/bin/env python
# -*- coding: utf-8 -*-
# author： Wentao Zheng
# datetime： 2024/3/4 21:13 
# ide： PyCharm
import os
import sys
import shlex
import platform
import subprocess
from pathlib import Path
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),'..')))

# 自定义库
from dynamic_scenes.socket_module import Client
from dynamic_scenes.observation import Observation

class KineticsModelStarter():
    def __init__(self, observation: Observation):
        initial_state = self._get_init_state_of_ego(observation)
        self.client = Client()
        self._write_temp_script(initial_state[0], initial_state[1], initial_state[2], initial_state[3])
        self._check_completed()
        pass

    @property
    def get_client(self):
        return self.client

    def _get_init_state_of_ego(self, observation:Observation):
        x = observation.vehicle_info['ego']['x']
        y = observation.vehicle_info['ego']['y']
        yaw = observation.vehicle_info['ego']['yaw_rad']
        v0 = observation.vehicle_info['ego']['v_mps']
        return x,y,yaw,v0

    def _write_temp_script(self, x:float, y:float, yaw:float, v0:float):
        # 编写一个tempScript.m脚本用于存储初始化信息
        current_path = Path(__file__).parent
        os_type = self._judge_platform()
        tempscript = current_path.parent / 'kinetic_model' / f'{os_type}'/'tempScript.m'
        with open(tempscript, 'w') as f:
            f.write(f"x0={x};\n")
            f.write(f"y0={y};\n")
            f.write(f"head={yaw};\n")
            f.write(f"v0={v0};\n")
            f.write("acc=0.0;\n")  # 初始加速度
            f.write("gear=2;\n")  # 初始档位：1-前进档；2-驻车档；3-倒车档
            f.write("yaw=0.0;\n")  # 初始前轮转角
            f.write("load('a_brake.mat');\n")
            f.write("load('a_thr.mat');\n")
            f.write("load('brake.mat');\n")
            f.write("load('thr.mat');\n")
            f.write("modelName='VehicleModel';\n")
            f.write("run('control_simulink_with_udp.m');\n")

        command = f"matlab -r \"run('{tempscript.as_posix()}')\""
        result = subprocess.Popen(shlex.split(command))
        return result

    def _check_completed(self):
        # Check whether the initialization is complete
        data, _ = self.client.client_receive_sock.recvfrom(1024)  # 假设信号很小，不需要大缓冲区
        if data.decode() == 'ready':
            print("MATLAB就绪，继续执行")

    def _judge_platform(self):
        os_type = platform.system()
        if os_type == "Windows":
            return 'win'
        elif os_type == "Linux" or os_type == "Darwin":
            return 'linux'
        else:
            print(f"不支持的操作系统: {os_type}")


import copy
import math
import time

def sign(x):
    if x>=0:return 1
    else:return -1

class Point2D:
    """2D平面的位姿点."""
    def __init__(self,x,y,yaw_rad:float=0.0):
        self.x = x
        self.y = y
        self.yaw = yaw_rad

class CollisionLookup:
    def __init__(self):
        self._cSize = 0.2  # 每个单元格的大小
        self._position_resolution = 1  # 车辆位置分辨率
        self._positions = self._position_resolution**2  # 车辆位置的总数
        self._ego_width = 5.5  # 主车宽度
        self._ego_length = 8.8  # 主车长度
        self._locationPoint2Head = 7.2  # 后轴中心到前轮的长度
        self._locationPoint2Rear = 1.6  # 后轴中心到后轮的距离
        self._bbSize = math.ceil((math.sqrt(self._ego_width**2 + (self._locationPoint2Head*2)**2) + 4) / self._cSize)  # 包含车的栅格数
        self._headings = 72  # 72个离散角度
        self._delt_heading_rad = 3.1415926*2/72
        #生成每个栅格的离散点位
        self._points = [Point2D(0.,0.) for i in range(self._positions)]
        for i in range(self._position_resolution):
            for j in range(self._position_resolution):
                pt = Point2D(1.0/self._position_resolution*j,\
                                1.0/self._position_resolution*i)
                self._points[i*self._position_resolution+j]=pt  # 维度为1*self.resolution**2
        self._c = Point2D(0.,0.) # 车辆中心
        self._p = [Point2D(0.,0.) for i in range(4)] #车辆的四个角
        self._np = [Point2D(0.,0.) for i in range(4)] #旋转后车辆的四个角
        self._lookup = [list() for i in range(self._positions*self._headings)] #表格
        for q in range(self._positions):
            theta = 0.0
            #计算车辆中心
            self._c.x = self._bbSize / 2.0 + self._points[q].x
            self._c.y = self._bbSize / 2.0 + self._points[q].y
            #计算车辆四个角位置
            self._p[0].x = self._c.x - self._locationPoint2Rear/self._cSize  # 6
            self._p[0].y = self._c.y - self._ego_width/2/self._cSize  # 10.5
            self._p[1].x = self._c.x - self._locationPoint2Rear/2/self._cSize  # 6
            self._p[1].y = self._c.y + self._ego_width/2/self._cSize  # 17.5
            self._p[2].x = self._c.x + self._locationPoint2Head/self._cSize
            self._p[2].y = self._c.y + self._ego_width/2/self._cSize
            self._p[3].x = self._c.x + self._locationPoint2Head/self._cSize
            self._p[3].y = self._c.y - self._ego_width/2/self._cSize

            for o in range(self._headings):
                self._cSpace = [False for i in range(self._bbSize**2)] #初始化C空间

                #旋转变换
                for i in range(4):
                    temp_x = self._p[i].x - self._c.x
                    temp_y = self._p[i].y - self._c.y
                    self._np[i].x = temp_x*math.cos(theta) - temp_y*math.sin(theta) + self._c.x
                    self._np[i].y = temp_x*math.sin(theta) + temp_y*math.cos(theta) + self._c.y
                theta+=self._delt_heading_rad

                #单元格顺时针遍历
                for i in range(4):
                    start = Point2D(0,0)
                    end = Point2D(0,0)
                    #起始点和终点设置
                    if i<3:
                        start = self._np[i]  # (6, 10.5)
                        end = self._np[i+1]  # (6, 17.5)
                    else:
                        start = self._np[i]
                        end = self._np[0]
                    
                    #x,y向下取整得到下标
                    x_id = math.floor(start.x)  # 6
                    y_id = math.floor(start.y)  # 10
                    self._cSpace[y_id*self._bbSize+x_id] = True #占用

                    #搜索起点到终点的方向向量
                    t = Point2D(end.x - start.x, end.y - start.y)  # (0, 7)
                    #下一步移动方向(按格点移动)
                    step_x = sign(t.x)  # 1
                    step_y = sign(t.y)  # 1
                    #下一步移动距离占总移动距离的比例(x,y方向)
                    t_delta_x = 1000
                    t_delta_y = 1000
                    if t.x!=0:t_delta_x = 1./abs(t.x)  # 一个格点占x移动距离的比例
                    if t.y!=0:t_delta_y = 1./abs(t.y)  # 一个格点占y移动距离的比例
                    t_max_x,t_max_y = 0.,0.
                    if step_x > 0:t_max_x = t_delta_x*(1-(start.x - math.trunc(start.x)))
                    else:t_max_x = t_delta_x*(start.x - math.trunc(start.x))
                    if step_y > 0:t_max_y = t_delta_y*(1-(start.y - math.trunc(start.y)))
                    else:t_max_y = t_delta_y*(start.y - math.trunc(start.y))

                    #边界占用栅格搜索
                    # 这里的优先移动的原则为:优先移动距离较远的方向（即移动的一格相较于该方向的距离来说，微不足道（影响较小）的方向）
                    # 确定方向后，要再确定移动后在该方向上离目标点更近了
                    while math.floor(end.x)!=x_id or math.floor(end.y)!=y_id:
                        #当x所占比例小,且移动后距离终点更近时,移动x
                        if t_max_x<t_max_y and abs(x_id+step_x-math.floor(end.x))<abs(x_id-math.floor(end.x)):
                            t_max_x += t_delta_x
                            x_id += step_x
                            self._cSpace[y_id*self._bbSize+x_id]=True
                        elif t_max_y<t_max_x and abs(y_id+step_y-math.floor(end.y))<abs(y_id-math.floor(end.y)):
                            t_max_y += t_delta_y
                            y_id += step_y
                            self._cSpace[y_id*self._bbSize+x_id]=True
                        elif 2 >= abs(x_id-math.floor(end.x)) + abs(y_id-math.floor(end.y)):  # 判断终点
                            if abs(x_id-math.floor(end.x)) > abs(y_id-math.floor(end.y)):
                                x_id += step_x
                                self._cSpace[y_id*self._bbSize+x_id]=True
                            else:
                                y_id += step_y
                                self._cSpace[y_id*self._bbSize+x_id]=True
                        else:
                            print("\n--->tie occured,please check for error in script\n")
                            break

                #填满图形
                hcross1,hcross2=0,0
                for i in range(self._bbSize):
                    inside = False
                     #寻找覆盖区域的起点和终点
                    for k in range(self._bbSize):
                        if self._cSpace[i*self._bbSize+k]==True and inside==False:
                            hcross1 = k
                            inside = True
                        elif self._cSpace[i*self._bbSize+k]==True and inside==True:
                            hcross2 = k
                    #如果没有找到覆盖区域
                    if inside==False:continue
                    #填充覆盖区域
                    for j in range(self._bbSize):
                        if j>=hcross1 and j<=hcross2:
                            self._cSpace[i*self._bbSize+j]=True
                
                #设置表格
                for i in range(self._bbSize):
                    for j in range(self._bbSize):
                        if self._cSpace[i*self._bbSize+j]:
                            self._lookup[q*self._headings+o].append(Point2D(j-math.floor(self._c.x),\
                                                                                i-math.floor(self._c.y))) # 相较于中心的坐标
                # #debug
                # for i in range(self._bbSize):
                #     print(" ")
                #     for j in range(self._bbSize):
                #         if self._cSpace[i*self._bbSize+j]==True:
                #             print("#",end='')
                #         else:
                #             print("-",end='')


    """
    @brief collisionDetection--碰撞检测
    输入:x,y——与地图原点的相对坐标,h——偏航角,image——二值化地图(observation['hdmaps_info']['image_mask'].image_ndarray)
    """
    def collisionDetection(self,x,y,h,image):
        start_time = time.perf_counter()
   
        x = x/0.1   #self._cSize
        y = y/0.1   #self._cSize
        X = math.floor(x)
        Y = math.floor(y)
        iX = math.floor((x-X)*self._position_resolution)
        iY = math.floor((y-Y)*self._position_resolution)
        while h>math.pi*2:
            h-=math.pi*2
        while h<0:
            h+=math.pi*2
        iH = math.floor(h/self._delt_heading_rad)
        idx = iY*self._position_resolution*self._headings + iX*self._headings + iH

        for i in range(len(self._lookup[idx])):
            cX = X+self._lookup[idx][i].x*2
            cY = Y+self._lookup[idx][i].y*2
            # if cX>=0 and cX<image.shape[1] and cY>=0 and cY<image.shape[0]:
            #     print(image[image.shape[0]-22000][image.shape[1]-8500])
            if image[cY][cX]==False:
                return True
        
        # flag = False
        # for j in range(13220,13120,-1):
        #     for i in range(7476,7576):
            
        #         # for k in self._lookup[idx]:
        #         #     if i == X+k.x and j == Y+k.y:
        #         #         flag = True
        #         # if flag == True:
        #         #     if image[j][i] == True:
        #         #         print("1",end='')
        #         #     else:
        #         #         print('0',end='')
        #         # else:
        #         #     print(' ',end='')
        #         # flag = False
        #         if image[image.shape[0]-j][i] == True:
        #             print("1",end='')
        #         else:
        #             print('0',end='')
        #     print('---\n')
        end_time = time.perf_counter()
        # print("碰撞检测时间:"+str(end_time-start_time))
        return False

if __name__ == "__main__":
    start_time = time.perf_counter()
    lookup = CollisionLookup()
    end_time = time.perf_counter()
    print("运行时间:"+str(end_time-start_time))
        



import json
from dynamic_scenes.observation import Observation
import pandas as pd
from functools import reduce
import copy


class DataRecord():
    def __init__(self):
        self.data = {}
        self.vehicle_column = ['x','y','yaw_rad','v_mps','acc_mpss','width','length','locationPoint2Head','locationPoint2Rear']

    def add_data(self,observation:Observation):
        """将输入的观察值进行存储
        """
        # 首先将已经建立了对应DataFrame的车辆名提取出来
        stored_vehicles = self.data.keys()

        # 提取observation对应的时刻
        t = observation.test_setting['t']

        # 遍历observation中的所有车辆
        for vehicle_name,values in observation.vehicle_info.items():
            # 如果vehi_name对应的车还没有建立DataFrame,则先建立
            if vehicle_name not in stored_vehicles:
                self._add_vehicle_frame(vehicle_name)

            ###### 每个车辆要保存的数据,注意和 self.vehicle_column 对应 #####
            values_saved={}
            values_saved= {'x':values['x'],
                           'y':values['y'],
                           'yaw_rad':values['yaw_rad'],
                           'v_mps':values['v_mps'],
                        #    'yawrate_radps':values['yawrate_radps'],#该值不需要储存
                           'acc_mpss':values['acc_mpss'],
                           'width':values['shape']['width'],
                           'length':values['shape']['length'],
                           'locationPoint2Head':values['shape']['locationPoint2Head'],
                           'locationPoint2Rear':values['shape']['locationPoint2Rear']
                            }
            ###### 每个车辆要保存的数据,注意和 self.vehicle_column 对应 end #####
            
            # 为t时刻的数据建立当前时刻的DataFrame
            sub_frame = pd.DataFrame(
                values_saved,
                columns=self.vehicle_column,
                index=[t]
            )
            # 修改列名,便于合并
            sub_frame.columns = list(self.data[vehicle_name].columns)
            # 将当前时刻的DataFrame加入车辆的DataFrame中
            # self.data[vehicle_name] = pd.concat([self.data[vehicle_name],sub_frame])

            # 检查sub_frame是否为空或全部为NA，如果不是，则进行合并
            if not sub_frame.empty and not sub_frame.isna().all().all():
                self.data[vehicle_name] = pd.concat([self.data[vehicle_name], sub_frame])

    def merge_frame(self) -> pd.DataFrame:
        """将存储的所有交通参与者的DataFrame,按照时间进行合并,返回完整的DataFrame

        """
        # 取出每辆车的DataFrame,组成列表
        vehicle_dataframe_group = [self.data[vehi_name]
                                   for vehi_name in list(self.data.keys())]
        # 返回合并后的DataFrame
        return reduce(lambda x,y:pd.merge(x,y,how="outer",left_index=True,right_index=True),vehicle_dataframe_group)

    def _add_vehicle_frame(self,vehicle_name:str):
        """为某一交通参与者创造对应的小DataFrame

        """
        self.data[vehicle_name] = pd.DataFrame(
            None,
            columns=[i + "_" + str(vehicle_name) for i in self.vehicle_column]
        )


class Recorder():
    """记录场景运行信息

    """

    def __init__(self):
        self.output_dir = ""
        self.file_name = ""
        self.data_record = DataRecord()

    def init(self,observation:Observation,output_dir:str,read_only:bool = False) -> None:
        if output_dir[-1] != "/":
            output_dir += "/"
        # 创建新的数据容器,目的是删除之前的数据
        self.data_record = DataRecord()
        self.output_dir = output_dir
        self.read_only = read_only
        self.record(observation)

    def record(self,observation:Observation) -> None:
        """记录当前时刻的运行信息,如果发现end status表示为测试结束,则写入文件.
        """
        # 记录当前时刻的数据
        self.data_record.add_data(observation)
        # 合并.
        data_output = copy.deepcopy(self.data_record.merge_frame())
        # 增加结束状态一列
        data_output.loc[:,'end'] = -1
        # 将最后一个值调整为观察值中的end,若-1则表示规控器问题导致测试不能顺利结束,若1则表示到达最大时间结束,若2则说明发生碰撞
        t = observation.test_setting['t']
        # 若超出地图范围,去掉最后一帧,防止评价体系失效
        if observation.test_setting['end'] == 1.5:
            data_output.drop(t,axis=0,inplace=True)
            data_output.iloc[-1,-1] = 1
        else:
            data_output.iloc[-1,-1] = observation.test_setting['end']
        # 若非只读模式,写入文件,每一步都会写入,自动覆盖上一步的文件
        if not self.read_only:
            data_output.to_csv(
                self.output_dir + observation.test_setting['scenario_name'] + "_result.csv")
        # 如果测试不是中途结束,即不是规控器自身的问题,那么会打印输出test_setting.
        if observation.test_setting['end'] != -1:
            show_test_setting = json.dumps(observation.test_setting['end'], indent=4, ensure_ascii=False)
            print(show_test_setting)


if __name__ == "__main__":
    data = DataRecord()
    observe = Observation()
    data.add_data(observe)
    print(data.data)



class Observation():
    """
    vehicle_info={
        "ego":{
            "x":,
            "y":,
            "yaw_rad":,
            "v_mps":,
            "yawrate_radps":,
            "acc_mpss":,
            "shape":{
				"vehicle_type":"MineTruck_NTE200",
				"length":13.4,
				"width":6.7,
				"height":6.9,
				"min_turn_radius":14.2,
				"locationPoint2Head":9.8,
				"locationPoint2Rear":3.8
			}
        },
        "0":{...},
        ...
    }
    test_setting = {
            "t":,
            "dt":,
            "max_t",
            "goal":{
                "x":[-1,-1,-1,-1],
                "y":[-1,-1,-1,-1]
            },
            "end":
        }
    """

    def __init__(self):
        self.vehicle_info = {
            "ego":{
                "x":-1,
                "y":-1,
                "yaw_rad":-1,
                "v_mps":-1,
                "yawrate_radps":-1,
                "acc_mpss":-1,
                "shape":{
                    "vehicle_type":"MineTruck_NTE200",
                    "length":13.4,
                    "width":6.7,
                    "height":6.9,
                    "min_turn_radius":14.2,
                    "locationPoint2Head":9.8,
                    "locationPoint2Rear":3.8
			    }
            },
        }
        self.hdmaps = {}
        self.test_setting = {
            "scenario_name":"name",
            "scenario_type":"intersection",
            "enter_loading_flag":False,
            "enter_loading_time":0.00,
            "t":0.0,
            "dt":0.1,
            "max_t":-1,
            "goal":{
                "x":[-1,-1,-1,-1],
                "y":[-1,-1,-1,-1],
                "head":None
            },
            "end":-1,
            "x_min":None,
            "x_max":None,
            "y_min":None,
            "y_max":None,
            "start_ego_info":None  
        }
        
    def format(self):
        return {
            "vehicle_info":self.vehicle_info,
            "test_setting":self.test_setting,
            "hdmaps_info":self.hdmaps,
        }

if __name__ == "__main__":
    pass



# encoding=utf-8
# Author：Wentao Zheng
# E-mail: swjtu_zwt@163.com
# developed time: 2024/2/8 12:17
import socket
import struct  # 解析simulink模型打包来的数据要用
import time
import platform
import subprocess

class Client():
    def __init__(self, Send_IP='127.0.0.1', Send_Port=25001, Receive_IP='127.0.0.1', Receive_Port=25000):
        self.send_ip = Send_IP
        self.send_port = Send_Port
        self.receive_ip = Receive_IP
        self.receive_port = Receive_Port
        self._build_client()

    def _build_client(self):
        # 发送端
        self.client_send_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # 接收端
        self.client_receive_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.client_receive_sock.bind((self.receive_ip, self.receive_port))

    def send_and_receive(self, gear, acceleration, steering_angle):
        try:
            message = struct.pack('>ddd', gear, acceleration, steering_angle)
            self.client_send_sock.sendto(message, (self.send_ip, self.send_port))
            print('控制量已发送，等到接收下一个状态量...')

            data, addr = self.client_receive_sock.recvfrom(1024)
            print('状态量已接收，等待计算下一个控制量...')
            if data:
                unpacked_data = struct.unpack('>dddd', data)
                return unpacked_data
        except Exception as e:
            print(f"通信出现问题！，具体原因为{e}")

    def close_sockets(self):
        self.client_send_sock.close()
        self.client_receive_sock.close()
        self.kill_matlab_processes()

    @staticmethod
    def kill_matlab_processes():
        # 获取当前操作系统类型
        os_type = platform.system()

        try:
            if os_type == "Windows":
                # 对于Windows系统，使用taskkill
                subprocess.run(["taskkill", "/F", "/IM", "MATLAB.exe"], check=True)
                print("所有MATLAB进程已被成功终止。")
            elif os_type == "Linux" or os_type == "Darwin":
                # 对于Linux和MacOS系统，使用pkill
                # MacOS系统也被视为类Unix系统，通常使用和Linux相同的命令
                subprocess.run(["pkill", "matlab"], check=True)
                print("所有MATLAB进程已被成功终止。")
            else:
                print(f"不支持的操作系统: {os_type}")
        except subprocess.CalledProcessError as e:
            print(f"终止MATLAB进程时发生错误：{e}")








# 内置库 
import math

# 第三方库
import numpy as np
import matplotlib.pyplot as plt
from typing import List,Dict,Set,Tuple,Union



def filterForYawBySinAndCos(angle_rad,Q_oneDim,R_oneDim,smooth_N1,smooth_N2):
    """ 对航向值的 sin cos 值 进行过滤
        
        Parameters
        ----------
        angle_rad :角度列表 ,单位 rad  # 列向量
        Q_oneDim :xxxx   
        R_oneDim :xxxx    
        smooth_N1 :xxxx 
        smooth_N2 :xxxx 

        Returns
        -------
        angle_rad_KF_smooth : 返回 滤波后的 angle ,单位 rad          
    """
    NTemp = angle_rad.shape[0]
    angle_sin = np.sin(angle_rad)
    angle_cos = np.cos(angle_rad)
    angle_rad_KF_smooth = np.zeros(NTemp)
    
    angle_sin_KF = KalmanFilter_FirstOrder(angle_sin,Q_oneDim,R_oneDim)
    angle_sin_KF_smooth = angle_sin_KF #不再进行移动平均滤波

    angle_cos_KF = KalmanFilter_FirstOrder(angle_cos,Q_oneDim,R_oneDim)
    angle_cos_KF_smooth = angle_cos_KF

    for i in range(NTemp):
        if angle_sin_KF_smooth[i] >= 0 and angle_cos_KF_smooth[i] >= 0: # 第一象限
            angle_rad_KF_smooth[i] = np.arccos(angle_cos_KF_smooth[i])
        elif angle_sin_KF_smooth[i] >= 0 and angle_cos_KF_smooth[i] < 0: # 第二象限
            angle_rad_KF_smooth[i] = np.arccos(angle_cos_KF_smooth[i])
        elif angle_sin_KF_smooth[i] < 0 and angle_cos_KF_smooth[i] < 0: # 第三象限
            angle_rad_KF_smooth[i] = 2* np.pi - np.arccos(angle_cos_KF_smooth[i])
        elif angle_sin_KF_smooth[i] < 0 and angle_cos_KF_smooth[i] >= 0: # 第四象限
            angle_rad_KF_smooth[i] = 2* np.pi - np.arccos(angle_cos_KF_smooth[i])
        else:
            # print("### log ### 一共 xxxxx")
            pass

    return angle_rad_KF_smooth


def KalmanFilter_FirstOrder(data,Q,R):
    """ 线性卡尔曼滤波,一维
    """
    N = len(data)
    KalmanGain = np.zeros(N)
    X_hat = np.zeros(N)
    X_hat_pre = np.zeros(N)
    P_covariance = np.zeros(N)
    P_covariance_pre = np.zeros(N)

    X_hat_pre[0] = data[0]
    P_covariance[0] = 1

    # 系统矩阵
    A = 1  # 状态转移矩阵
    H = 1  # 测量矩阵
    I = np.eye(1)

    for k in range(N-1):
        # 卡尔曼滤波过程
        X_hat_pre[k+1] = A * X_hat[k]  # 1) 预测:先验状态估计值
        P_covariance_pre[k+1] = A * P_covariance[k] * A + Q  # 2) 预测:先验预测协方差矩阵

        KalmanGain[k+1] = P_covariance_pre[k+1] * H /(H * P_covariance_pre[k+1] * H + R)  # 3) 校正:计算Kalman增益矩阵

        X_hat[k+1] = X_hat_pre[k+1] + KalmanGain[k+1] * (data[k+1] - H * X_hat_pre[k+1])  # 4) 校正:后验状态估计更新
        P_covariance[k+1] = (I - KalmanGain[k+1] * H) * P_covariance_pre[k+1]  # 5) 校正:误差协方差矩阵更新

    return X_hat


def KalmanFilter_linear(measurements,initial_state,
                        transition_matrix,observation_matrix,
                        process_noise_covariance,measurement_noise_covariance):
    """ 有问题,未调好
    """
    num_measurements = len(measurements)
    # num_states = initial_state.shape[0]
    num_states = 1

    # 初始化状态估计和协方差矩阵
    state_estimation = np.zeros((num_states,num_measurements))
    state_covariance = np.zeros((num_states,num_states,num_measurements))
    
    # 初始化先验状态估计和协方差矩阵
    predicted_state_estimation = np.zeros((num_states,num_measurements))
    predicted_state_covariance = np.zeros((num_states,num_states,num_measurements))

    # 初始状态估计
    state_estimation[:,0] = initial_state
    state_covariance[:,:,0] = np.eye(num_states)

    for t in range(1,num_measurements):
        # 预测步骤
        predicted_state_estimation[:,t] = transition_matrix @ state_estimation[:,t-1]
        predicted_state_covariance[:,:,t] = transition_matrix @ state_covariance[:,:,t-1] @ transition_matrix.T + process_noise_covariance
        
        # 更新步骤
        kalman_gain = predicted_state_covariance[:,:,t] @ observation_matrix.T @ np.linalg.inv(observation_matrix @ \
            predicted_state_covariance[:,:,t] @ observation_matrix.T + measurement_noise_covariance)
        state_estimation[:,t] = predicted_state_estimation[:,t] + kalman_gain @ (measurements[t] - observation_matrix @ \
            predicted_state_estimation[:,t])
        state_covariance[:,:,t] = (np.eye(num_states) - kalman_gain @ observation_matrix) @ predicted_state_covariance[:,:,t]
    
    return state_estimation


def smooth_data(data,window_size):
    """This function applies a simple moving average filter to the data by convolving it with a window of ones.
        该函数通过将数据与一个1的窗口进行卷积,对数据应用一个简单的移动平均过滤器.
    """
    window = np.ones(window_size) / window_size
    smoothed_data = np.convolve(data,window,mode='valid')
   
    return smoothed_data


def surrounding_id (observation,ego_x,ego_y,r):
    '''该函数输入车辆的XY坐标以及检查范围,输出车辆范围内的其他车辆ID
    '''
    id_veh = []
    for key,value in observation['vehicle_info'].items():
        if key != 'ego':
            dis = math.hypot(abs(observation['vehicle_info'][key]['x']-ego_x),abs(observation['vehicle_info'][key]['y']-ego_y))
            if dis < r:
                id_veh.append(key)
    return id_veh


# ! 以上暂时无用

      
def find_nearest_point_index(X:float,Y:float,path_points_XY:np.array) ->int:
    """find id of nearest a path point

    Args:
        X (float):匹配点x坐标;
        Y (float):匹配点y坐标;
        path_points_XY (np.array):路径点序列;保证前两列为x y坐标值即可

    Returns:
        int:最近的路径点id
    """
    min_distance = math.inf
    nearest_id = -1
    for i,point in enumerate(path_points_XY):
        point_x,point_y,_ = point
        distance = math.sqrt((X - point_x) ** 2 + (Y - point_y) ** 2)
        if distance < min_distance:
            min_distance = distance
            nearest_id = i
    return nearest_id
    
    
def find_preview_point_index(X:float,Y:float,preview_dis:float,path_points_XY:np.array) ->int:
    """按照预瞄距离preview_dis查找参考路径点索引

    Args:
        X (float):匹配点x坐标
        Y (float):匹配点y坐标
        preview_dis (float):向前方向参考路径预瞄距离
        path_points_XY (np.array):路径点序列;保证前两列为x y坐标值即可

    Returns:
        int:预瞄路径点的id
    """
    nearest_id = find_nearest_point_index(X,Y,path_points_XY)
    preview_id = nearest_id
    accumulated_distance = 0.0
    for i in range(nearest_id,len(path_points_XY-1)):
        distance = math.sqrt((path_points_XY[i+1][0] - path_points_XY[i][0])**2 + (path_points_XY[i+1][1]- path_points_XY[i][1])**2)
        accumulated_distance += distance  # 使用累加的距离
        if accumulated_distance > preview_dis:   
            break #如果预瞄距离到达最后id,就给出最后的id
        preview_id = i+1
        
    return preview_id


def find_preview_point_index_over_end(X:float,Y:float,preview_dis:float,path_points_XY:np.array) -> Union[int,bool]:
    """按照预瞄距离preview_dis查找参考路径点索引,允许超出界限end一定程度

    Args:
        X (float):匹配点x坐标
        Y (float):匹配点y坐标
        preview_dis (float):向前方向参考路径预瞄距离
        path_points_XY (np.array):路径点序列;保证前两列为x y坐标值即可

    Returns:
        int:预瞄路径点的id
    """
    over_index_distance_ratio = 0.2 # !索引超界;设置预瞄距离可以超过 20% [可配置参数项]
    flag_preview_dis_over_end = False #初始值
    
    nearest_id = find_nearest_point_index(X,Y,path_points_XY)
    preview_id = nearest_id
    accumulated_distance = 0.0
    for i in range(nearest_id,len(path_points_XY-1)):
        if i+1 >= len(path_points_XY)-1:###### 索引超界判断,并处理 #####
            over_index_distance_ratio_temp=(preview_dis - accumulated_distance)/accumulated_distance
            if over_index_distance_ratio_temp<over_index_distance_ratio:
                flag_preview_dis_over_end = False
                preview_id = i
                break
            else:
                flag_preview_dis_over_end = True
                preview_id = i
                break
            ###### 索引超界判断,并处理 end#####
        else:
            distance = math.sqrt((path_points_XY[i+1][0] - path_points_XY[i][0])**2 + (path_points_XY[i+1][1]- path_points_XY[i][1])**2)
            accumulated_distance += distance  # 使用累加的距离
            if accumulated_distance > preview_dis:   
                break #如果预瞄距离到达最后id,就给出最后的id
            preview_id = i+1
        
    return preview_id,flag_preview_dis_over_end


def compute_two_pose_error(X1:float,Y1:float,Yaw1:float,X2:float,Y2:float,Yaw2:float)-> Tuple[float,float,float]:
    """计算当前位姿点 pose1 = (X1,Y1,Yaw1) 到目标位姿点 pose2 = (X2,Y2,Yaw2) 的误差项(位姿关系)

    Args:
        X1 (float):单位m
        Y1 (float):单位m
        Yaw1 (float):单位 rad , 东偏北,范围(0,2*pi]
        X2 (float):单位m
        Y2 (float):单位m
        Yaw2 (float):单位 rad , 东偏北,范围(0,2*pi]

    Returns:
        Tuple[float,float,float]:lateral_error 横向误差(左正右负),pose1在pose2左侧为正;
                                    longitudinal_error 纵向误差(前正后负),pose1在pose2前方为正;
                                    heading_error_rad 航向误差(顺正逆负) ,pose1相对于pose2顺时针旋转为正;
    """
    dx = X1 - X2
    dy = Y1 - Y2

    lateral_error = math.cos(Yaw2) * dy - math.sin(Yaw2) * dx
    longitudinal_error = math.sin(Yaw2) * dy + math.cos(Yaw2) * dx

    heading_error_rad_temp = Yaw1 - Yaw2 # Calculate heading error in radians
    # Normalize heading error to the range [-pi,pi]
    heading_error_rad = (heading_error_rad_temp + np.pi) % (2 * np.pi) - np.pi

    return lateral_error,longitudinal_error,heading_error_rad


def calculate_longitudinal_distance(point,dubinspose):
    """计算当前车辆定位点 point = (x2,y2) 到位姿点 dubinspose = (x1,y1,yaw1) 的纵向距离,"""
    # ! 未验证
    x1,y1,yaw1 = dubinspose
    x2,y2 = point
    
    # 计算法线方向向量
    normal_vector = (math.cos(yaw1 + math.pi/2),math.sin(yaw1 + math.pi/2))
    
    # 计算向量差
    vector_diff = (x2 - x1,y2 - y1)
    
    # 计算纵向距离
    distance = vector_diff[0] * normal_vector[0] + vector_diff[1] * normal_vector[1]
    
    return distance
 
 
def calculate_vehicle_corners(length,width,locationPoint2Head,locationPoint2Rear,x,y,yaw_rad):
    """根据提供的车辆形状参数和位置信息,计算车辆的四个顶点坐标. 

    Args:
        length (_type_): 车辆长度
        width (_type_): 车辆宽度
        locationPoint2Head (_type_): 车辆定位中心距离车头长度
        locationPoint2Rear (_type_): 车辆定位中心距离车尾部长度
        x (_type_): 车辆定位点坐标x
        y (_type_): 车辆定位点坐标y
        yaw_rad (_type_): 车辆航向 

    Returns:
        float: 车辆四个角点坐标(x,y).
    """
    
    # 计算车辆中心点
    center = np.array([x,y])
    
    # 计算方向矢量
    direction = np.array([np.cos(yaw_rad),np.sin(yaw_rad)])
    perpendicular_direction = np.array([-np.sin(yaw_rad),np.cos(yaw_rad)])
    
    # 计算车辆前后中心点
    front_center = center + locationPoint2Head * direction
    rear_center = center - locationPoint2Rear * direction
    
    # 计算车辆半长和半宽矢量
    half_length_vec = 0.5 * length * direction
    half_width_vec = 0.5 * width * perpendicular_direction
    
    # 计算车辆四个角点
    front_left_corner = front_center + half_width_vec
    front_right_corner = front_center - half_width_vec
    rear_left_corner = rear_center + half_width_vec
    rear_right_corner = rear_center - half_width_vec

    return front_left_corner.tolist(),front_right_corner.tolist(),rear_left_corner.tolist(),rear_right_corner.tolist()


def is_inside_polygon(x:float,y:float,polygon:Dict[str,list]) -> bool:
    """使用射线法判断点 point(x,y) 是否在多边形内部.
    Args:
        x (float):x坐标
        y (float):y坐标
        polygon (Dict[str,list]):描述多边形的字典,例如:
            {'x':[895.3440941,890.0361717,907.9749535,913.2828758],
             'y':[2314.994762,2325.517345,2334.566173,2324.04359]}
    Returns:
        bool:如果点在多边形内部返回True,否则返回False
    """
    # 初始化交点数量为0
    intersect_count = 0
    
    num_points = len(polygon['x'])
    for i in range(num_points):
        # 为每一个线段提取开始点p1和结束点p2
        p1 = (polygon['x'][i],polygon['y'][i])
        p2 = (polygon['x'][(i + 1) % num_points],polygon['y'][(i + 1) % num_points])  # 取%保证可以回到顶点

        # 确保 p1 的 y 值小于 p2 的 y 值
        if p1[1] > p2[1]:
            p1,p2 = p2,p1
        
        # 判断水平射线是否与线段有交点
        if p1[1] <= y < p2[1]:
            x_intersect = (p2[0] - p1[0]) * (y - p1[1]) / (p2[1] - p1[1]) + p1[0]
            if x < x_intersect:
                intersect_count += 1
    
    # 如果交点数量为奇数,返回True,否则返回False
    return intersect_count % 2 == 1


 
class PolynomialInterpolation:
    """笛卡尔坐标系下五次多项式插值."""
    def __init__(self,pt1,pt2,v1,v2,a1,a2,delta_t=0.1):
        self.pt1 = pt1  # 起始点位置
        self.pt2 = pt2  # 终止点位置
        self.v1 = v1    # 起始点速度
        self.v2 = v2    # 终止点速度
        self.a1 = a1    # 起始点加速度
        self.a2 = a2    # 终止点加速度
        self.delta_t = delta_t  # 时间步长


    def quintic_polynomial_interpolation(self,t0=0,t1=5.0) ->np.array:
        """Quintic polynomial curve interpolation   五次多项式
            起点时间 t0 =0; 终点时间t1 =5 #多项式曲线固定时长 5s
        """
        # t0 = 0
        # t1 = 5
        # T = np.array([[t0**5,  t0**4,   t0**3,   t0**2,t0,  1],
        #               [5*t0**4,4*t0**3, t0**2,   2*t0,1,   0],
        #               [20*t0**3,12*t0**2,6*t0,    2,   0,   0],
        #               [t1**5,  t1**4,   t1**3,   t1**2,t1,  1],
        #               [5*t1**4,4*t1**3, 3*t1**2, 2*t1,1,   0],
        #               [20*t1**3,12*t1**2,6*t1,    2,   0,   0]])
        T=np.array([[0,      0,       0,       0,       0,   1],
                    [0,      0,       0,       0,       1,   0],
                    [0,      0,       0,       2,       0,   0],
                    [t1**5,  t1**4,   t1**3,   t1**2,   t1,  1],
                    [5*t1**4,4*t1**3, 3*t1**2, 2*t1,    1,   0],
                    [20*t1**3,12*t1**2,6*t1,    2,       0,   0]]) #T矩阵
        T_inv=np.linalg.inv(T)
        X = np.array([[self.pt1[0]],[self.v1[0]],[self.a1[0]],[self.pt2[0]],[self.v2[0]],[self.a2[0]]]) #X矩阵
        Y = np.array([[self.pt1[1]],[self.v1[1]],[self.a1[1]],[self.pt2[1]],[self.v2[1]],[self.a2[1]]]) #Y矩阵
        A = np.dot(T_inv,X)
        B = np.dot(T_inv,Y)

        # num = int(t1/self.delta_t)
        # 将时间从t0到t1离散化,获得离散时刻的轨迹坐标
        t = np.arange(t0,t1+self.delta_t,self.delta_t)
        interpolation_trajectory = np.zeros((len(t),4))  # 1-4列分别存放x,y,vx,vy
        for i in range(len(t)):
            # x位置坐标
            interpolation_trajectory[i,0] = np.dot([t[i]**5,t[i]**4,t[i]**3,t[i]**2,t[i],1],A)
            # y位置坐标
            interpolation_trajectory[i,1] = np.dot([t[i]**5,t[i]**4,t[i]**3,t[i]**2,t[i],1],B)
            # x速度
            interpolation_trajectory[i,2] = np.dot([5*t[i]**4,4*t[i]**3,3*t[i]**2,2*t[i],1,0],A)
            # y速度
            interpolation_trajectory[i,3] = np.dot([5*t[i]**4,4*t[i]**3,3*t[i]**2,2*t[i],1,0],B)

        return interpolation_trajectory


    def quintic_polynomial_interpolation_2(self,t1=5.0):
        """Quintic polynomial curve interpolation   五次多项式
            起点时间 t0 =0; 终点时间t1 =5 #多项式曲线固定时长 5s
            @吴佳琪 该部分有问题;
        """
        # t0 = 0
        # t1 = 5
        # T = np.array([[t0**5,  t0**4,   t0**3,   t0**2,t0,  1],
        #               [5*t0**4,4*t0**3, t0**2,   2*t0,1,   0],
        #               [20*t0**3,12*t0**2,6*t0,    2,   0,   0],
        #               [t1**5,  t1**4,   t1**3,   t1**2,t1,  1],
        #               [5*t1**4,4*t1**3, 3*t1**2, 2*t1,1,   0],
        #               [20*t1**3,12*t1**2,6*t1,    2,   0,   0]])
        T=np.array([[0,      0,       0,       0,       0,   1],
                    [0,      0,       0,       0,       1,   0],
                    [0,      0,       0,       2,       0,   0],
                    [t1**5,  t1**4,   t1**3,   t1**2,   t1,  1],
                    [5*t1**4,4*t1**3, 3*t1**2, 2*t1,    1,   0],
                    [20*t1**3,12*t1**2,6*t1,    2,       0,   0]]) #T矩阵
        T_inv=np.linalg.inv(T)
        X = np.array([[self.pt1[0]],[self.v1[0]],[self.a1[0]],[self.pt2[0]],[self.v2[0]],[self.a2[0]]]) #X矩阵
        Y = np.array([[self.pt1[1]],[self.v1[1]],[self.a1[1]],[self.pt2[1]],[self.v2[1]],[self.a2[1]]]) #Y矩阵
        A = np.dot(T_inv,X)
        B = np.dot(T_inv,Y)

        interpolation_trajectory = []
        num = int(t1/self.delta_t)
        for i in range(1,num+1):#1,2,3,...,50
            dt = i * self.delta_t
            T_dt = np.array([[0,      0,       0,       0,       0,   1],
                             [0,      0,       0,       0,       1,   0],
                             [0,      0,       0,       2,       0,   0],
                             [dt**5,  dt**4,   dt**3,   dt**2,   dt,  1],
                             [5*dt**4,4*dt**3, 3*dt**2, 2*dt,    1,   0],
                             [20*dt**3,12*dt**2,6*dt,    2,       0,   0]]) #T矩阵
            X_dt = np.dot(T_dt,X)
            Y_dt = np.dot(T_dt,Y)
            x_dt = X_dt[3][0]
            y_dt = Y_dt[3][0]
            v_dt = X_dt[4][0]
            point = [x_dt,y_dt,v_dt]
            interpolation_trajectory.append(point)

        return interpolation_trajectory


    # TODO 三次多项式曲线生成函数
    def cubic_polynomial_interpolation_2(self,t1=5.0):
        pass
    


if __name__ == "__main__":
    """测试代码"""
    pt1 = [0,-1.75]    # 起始点位置
    pt2 = [20,1.75]   # 终止点位置
    v1 = [5,0]     # 起始点速度
    v2 = [5,0]     # 终止点速度
    a1 = [0,0]     # 起始点加速度
    a2 = [0,0]     # 终止点加速度

    interpolator = PolynomialInterpolation(pt1,pt2,v1,v2,a1,a2)
    interpolation_trajectory = interpolator.quintic_polynomial_interpolation()

    # 提取插值结果的 x 和 y 坐标
    x_values = [point[0] for point in interpolation_trajectory]
    y_values = [point[1] for point in interpolation_trajectory]

    # 绘制插值路径
    plt.plot(x_values,y_values,'r--',linewidth=1.5,marker='.')
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.title('Quintic Polynomial Interpolation trajectory')
    plt.grid(True)
    plt.show()
    
    a=1 


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# author： Wentao Zheng
# datetime： 2023/11/17 14:47 
# ide： PyCharm

from setuptools import setup, find_packages

with open('./onsite-mine/setup/requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='onsite-mine',  # 包名
    version='0.1.1',  # 版本
    description="The simulation environment for unstructured road replay testing in onsite competitions.",  # 包简介
    long_description=open('README.md').read(),  # 读取文件中介绍包的详细内容
    include_package_data=True,  # 是否允许上传资源文件
    author='Zheng Wentao',  # 作者
    author_email='swjtu_zwt@163.com',  # 作者邮件
    maintainer='ZhengWentao',  # 维护者
    maintainer_email='swjtu_zwt@163.com',  # 维护者邮件
    license='MIT License',  # 协议
    url='https://github.com/WentaoZheng945/onsite_mine',  # github或者自己的网站地址
    packages=find_packages(),  # 用setuptools找到项目所有有关的包列表
    package_data={
            # 如果你的包名是 "your_package"，并且MATLAB脚本和模型位于一个名为 "matlab_scripts" 的目录中
            'onsite-mine': ['kinetic_model/win/*.mat',
                            'kinetic_model/win/*.m',
                            'kinetic_model/win/*.slxp',
                            'kinetic_model/win/*.slx',
                            'kinetic_model/linux/*.mat',
                            'kinetic_model/linux/*.m',
                            'kinetic_model/linux/*.slxp',
                            'kinetic_model/linux/*.slx',
                            ],
        },
    classifiers=[
        'Development Status :: 3 - Alpha',  # 包的开发状态，
        # Alpha 表示软件处于开发的早期阶段，可能仍然不够稳定，可能会有较大的变化。其他可能的值包括 Beta（测试阶段）、Production/Stable（稳定版）等。
        'Intended Audience :: Developers',  # 预期用户群
        # 这里指定了该包的预期用户为开发者。其他可能的值包括 End Users（终端用户）、Science/Research（科研人员）等。
        'Topic :: Software Development :: Testing',  # 主题和领域。
        # 解释：这里指定了该包与软件开发和测试相关。其他可能的值包括 Web（Web 开发）、Database（数据库相关）等
        'License :: OSI Approved :: MIT License',  # 许可协议
        'Programming Language :: Python :: 3',  # 设置编写时的python版本
    ],
    python_requires='>=3.9',  # 设置python版本要求
    install_requires=requirements,  # 安装所需要的库
    # entry_points={
    #     'console_scripts': [
    #         ''],
    # },  # 设置命令行工具(可不使用就可以注释掉)

)


# 内置库 
import os
import sys
import shlex
import subprocess
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),'..')))

# 第三方库
from typing import Dict,List,Tuple,Optional,Union

# 自定义库
from dynamic_scenes.observation import Observation
from dynamic_scenes.controller import Controller
from dynamic_scenes.recorder import Recorder
from dynamic_scenes.visualizer import Visualizer
from dynamic_scenes.lookup import CollisionLookup
from dynamic_scenes.kinetics_model import KineticsModelStarter




class Env():
    """仿真环境读取及迭代过程,simulation"""
    def __init__(self):
        self.controller = Controller()
        self.recorder = Recorder()
        self.visualizer = Visualizer()


    def make(self,scenario:dict,collision_lookup:CollisionLookup,read_only=False, save_img_path='',kinetics_mode='simple') -> Tuple:
        """第一次进入读取环境信息.

        Args:
            scenario (dict): 动态场景输入信息.
            collision_lookup (CollisionLookup): 用于 ego车与(栅格化mask)边界 进行碰撞检测的预置数据.
            read_only (bool, optional): _description_. Defaults to False.

        Returns:
            Observation: 当前时刻环境观察结果;
            traj:全局的背景车辆轨迹数据;
        """
        observation,traj = self.controller.init(scenario,collision_lookup,kinetics_mode)
        if kinetics_mode == "complex":
            Starter = KineticsModelStarter(observation)
            self.client = Starter.get_client
        elif kinetics_mode == "simple":
            self.client = None
        else:
            raise ValueError("暂不提供这种动力学模式，请在simple和complex中选择模式！")
        self.recorder.init(observation,scenario['file_info']['dir_outputs'],read_only)
        self.visualizer.init(observation,
                             scenario['test_settings']['visualize'],
                             scenario['test_settings']['save_fig_whitout_show'],
                             img_save_path=save_img_path) # 此处通过查看配置参数,True,设置运行过程中可视化打开;
        
        return observation.format(),traj,self.client


    def step(self,action:Tuple[float,float,int],traj_future:Dict,observation_last:Observation,traj:Dict,collision_lookup:CollisionLookup) -> Observation:
        """迭代过程"""
        observation = self.controller.step(action,collision_lookup,self.client)  # 使用车辆运动学模型单步更新场景;
        self.recorder.record(observation)
        # self.visualizer.update(observation)  
        self.visualizer.update(observation,traj_future,observation_last,traj)# 更新场景后,使用更新的ego车辆位置进行可视化;【CZF】添加预测轨迹+真实轨迹的比较
        
        return observation.format()





if __name__ == "__main__":
    import time
    demo_input_dir = r"demo/demo_inputs"
    demo_ouput_dir = r"demo/demo_outputs"
    tic = time.time()
    env = Env()

    from dynamic_scenes.scenarioOrganizer import ScenarioOrganizer
    # 实例化场景管理模块(ScenairoOrganizer)和场景测试模块(Env)
    so = ScenarioOrganizer()
    # 根据配置文件config.py装载场景,指定输入文件夹即可,会自动检索配置文件
    so.load(demo_input_dir,demo_ouput_dir)
    num_scenario = len(so.scenario_list)
    for i in range(num_scenario):
        scenario_to_test = so.next()
        print(scenario_to_test)
        observation = env.make(scenario_to_test,demo_ouput_dir,visilize=True)
        while observation.test_setting['end'] == -1:
            observation = env.step([-1,0])
            # print(observation.vehicle_info['ego'])
    toc = time.time()
    print(toc - tic)



# Code written by Chen Zhifa(陈志发,chenzhifa@buaa.edu.cn),2023.
# 内置库 
import sys
import os
import json
import math
from typing import Dict,List,Tuple,Optional,Union

# 第三方库
import numpy as np
import descartes
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from matplotlib.colors import to_rgba
from shapely import affinity
from shapely.geometry import Polygon,LineString,Point,box

# # 添加必要的路径到sys.path
def add_path_to_sys(target:str):
    abs_path = os.path.abspath(target)
    if abs_path not in sys.path:
        sys.path.append(abs_path)
dir_current_file = os.path.dirname(__file__)  # devkit/onsite_mine_code/map_expansion
dir_parent_1 = os.path.dirname(dir_current_file) # devkit/onsite_mine_code
dir_parent_2 = os.path.dirname(dir_parent_1) # devkit 
add_path_to_sys(dir_current_file)
add_path_to_sys(dir_parent_1)


# 自定义库
from map_expansion.bit_map import BitMap
import common.utils as utils

# Define a map geometry type for polygons and lines.
Geometry = Union[Polygon,LineString]
locations = ['jiangxi_jiangtong','guangdong_dapai', 'anhui_hailuo']



class TgScenesMap:
    """用于从语义地图中查询和检索多个图层信息.
    每张局部地图 有精准的原点(西南角,[lat,lon]):
    我们使用与谷歌Maps/Earth相同的WGS 84 Web Mercator (EPSG:3857)投影."CRS_name":"GCS_WGS_1984_UTM_Zone_50N",
    """
    def __init__(self,
                 dataroot:str='/home/czf/project_czf/20231010_onsite_mine/datasets',
                 location:str = 'jiangxi_jiangtong'):
        """加载 HD MAP 的多个图层数据,并初始化explorer class.

        Args:该函数接受两个参数
            dataroot (str, optional): datasets根目录. Defaults to '/home/czf/project_czf/20231010_onsite_mine/datasets'.
            location (str, optional): 具体矿区位置的名称,当前有以下可选 :`jiangxi_jiangtong`,`guangdong_dapai`. Defaults to 'jiangxi_jiangtong'.

        Raises:
            Exception: xxxx_semantic_map.json version is error.
        """
       
        assert location in locations,'Error:地图名字错误:%s!' % location

        self.dataroot = dataroot
        self.location = location
        
        self.geometric_layers = ['polygon','node','node_block']
        self.other_layers = ['dubins_pose','road_block'] 
        self.non_geometric_line_layers = ['reference_path','borderline']
        self.non_geometric_polygon_layers = ['road','intersection','loading_area','unloading_area']
        
        self.layer_names = self.geometric_layers + self.non_geometric_polygon_layers +self.other_layers 

        # 加载高精度矢量语义地图
        semantic_map_hashes = {
            "jiangxi_jiangtong":'jiangxi_jiangtong_semantic_map',
            "guangdong_dapai":'guangdong_dapai_semantic_map',
            "anhui_hailuo": 'anhui_hailuo_semantic_map'
        }
        semantic_map_hash = semantic_map_hashes[self.location]
        dir_semantic_map =  os.path.join(self.dataroot, "semantic_map" ,semantic_map_hash+'.json')
        with open(dir_semantic_map,'r') as f:
            self.semantic_map_json = json.load(f)

        # check the map version.
        if 'version' in self.semantic_map_json:
            self.version = self.semantic_map_json['version']
        else:
            self.version = '1.5'
        if self.version != '1.5':
            raise Exception('Error:只针对地图版本为 map version 1.5! 你的地图版本是',self.version)
        
        # load json data
        self._load_layers() # 加载各个图层 
        self._make_token2ind() # token标识符转化为id: 0,1,2, ...... ;
        

        self.explorer = TgScenesMapExplorer(self)  # 确定横纵比


    def _load_layer(self,layer_name:str) -> List[dict]:
        """Returns a list of info corresponding to the layer name. 返回各个图层 层名 对应的列表.

        Args:
            layer_name (str): Name of the layer that will be loaded.

        Returns:
            List[dict]: A list of info corresponding to a layer.
        """
        
        return self.semantic_map_json[layer_name]


    def _load_layers(self) -> None:
        """ Loads each available layer. 
        """
        self.node = self._load_layer('node')
        self.node_block = self._load_layer('node_block')
        self.polygon = self._load_layer('polygon')
        self.road= self._load_layer('road')
        self.intersection= self._load_layer('intersection')
        self.loading_area= self._load_layer('loading_area')
        self.unloading_area= self._load_layer('unloading_area')
        self.road_block = self._load_layer('road_block')
        self.dubins_pose = self._load_layer('dubins_pose')
        self.reference_path = self._load_layer('reference_path')
        self.borderline = self._load_layer('borderline')   
        

    def _make_token2ind(self) -> None:
        """ Store the mapping from token to layer index for each layer.
        token2ind = {
            "node": {"node-0": 0,
                     "node-1": 1
                    },
            "node_block": {"nodeblock-0": 0,
                           "nodeblock-1": 1,
                          },
        }
        """
        self.token2ind = dict()
        for layer_name in self.layer_names:
            # if layer_name == 'loading_area':  # only one
            #     continue
            self.token2ind[layer_name] = dict()
            for index,member in enumerate(getattr(self, layer_name)):
                self.token2ind[layer_name][member['token']] = index  # as: token2ind['node']['node_0'] = 0


    def get(self, layer_name: str, token: str) -> dict:
        """
        Returns a record from the layer in constant runtime.在固定的运行时间内从层中返回一条记录.
        :param layer_name: Name of the layer that we are interested in.
        :param token: Token of the record.
        :return: A single layer record.返回对应的dict
        """
        assert layer_name in self.layer_names, "Layer {} not found".format(layer_name)

        return getattr(self, layer_name)[self.getind(layer_name, token)]
    
    
    def getind(self, layer_name: str, token: str) -> int:
        """
        This returns the index of the record in a layer in constant runtime. .在固定的运行时间内从层中返回一条记录的id.
        :param layer_name: Name of the layer we are interested in.
        :param token: Token of the record.
        :return: The index of the record in the layer, layer is an array.返回对应层给定name时的索引
        """
        return self.token2ind[layer_name][token]


    def render_layers(self,
                      layer_names:List[str],
                      alpha:float = 0.5,
                      figsize:Union[None,float,Tuple[float,float]] = None,
                      tokens:List[str] = None,
                      bitmap:Optional[BitMap] = None) -> Tuple[Figure,Axes]:
        """
        Render a list of layer names.
        :param layer_names:A list of layer names.
        :param alpha:The opacity of each layer that gets rendered.
        :param figsize:Size of the whole figure.
        :param tokens:Optional list of tokens to render. None means all tokens are rendered.
        :param bitmap:Optional BitMap object to render below the other map layers.
        :return:The matplotlib figure and axes of the rendered layers.
        """
        return self.explorer.render_layers(layer_names,alpha,
                                           figsize=figsize,tokens=tokens,bitmap=bitmap)

    def render_map_patch(self,
                         box_coords:Tuple[float,float,float,float],
                         layer_names:List[str] = None,
                         alpha:float = 0.5,
                         figsize:Tuple[int,int] = (15,15),
                         render_egoposes_range:bool = True,
                         render_legend:bool = True,
                         bitmap:Optional[BitMap] = None,
                         fig:plt.figure = None) -> Tuple[Figure,Axes]:
        """
        Renders a rectangular patch specified by `box_coords`. By default renders all layers.
        :param box_coords:The rectangular patch coordinates (x_min,y_min,x_max,y_max).
        :param layer_names:All the non geometric layers that we want to render.
        :param alpha:The opacity of each layer.
        :param figsize:Size of the whole figure.
        :param render_egoposes_range:Whether to render a rectangle around all ego poses.
        :param render_legend:Whether to render the legend of map layers.
        :param bitmap:Optional BitMap object to render below the other map layers.
        :return:The matplotlib figure and axes of the rendered layers.
        """
        return self.explorer.render_map_patch(box_coords,layer_names=layer_names,alpha=alpha,figsize=figsize,
                                              render_egoposes_range=render_egoposes_range,
                                              render_legend=render_legend,bitmap=bitmap,fig = fig)

    def render_bitmap_mask(self,
                            patch_box_meter:Tuple[float,float,float,float],
                            patch_angle:float,
                            layer_names:List[str] = None,
                            figsize:Tuple[int,int] = (15,15),
                            n_row:int = 2) -> Tuple[Figure,List[Axes]]:
        """
        Render mask黑白掩码图像.
        :param Patch box :定义显示图框(补丁框)大小 [x_center,y_center,height,width],UTM坐标系,单位:米.
        :param patch_angle: 显示图框(补丁框)的方向角,单位:deg,正北为0 deg.
        :param layer_names:['intersection','road']或者 None
        :param canvas_size_pixel:Size of the outputs mask (h,w).
        :param figsize:Size of the figure.
        :param n_row:Number of rows with plots. 绘图的行数
        :return:The matplotlib figure and a list of axes of the rendered layers.
        """
        # 计算 patch_box_meter对应像素的宽*高,单位pix
        canvas_size_pixel = np.array((patch_box_meter[2],patch_box_meter[3])) * self.scale_PixelPerMeter
        canvas_size_pixel = tuple(np.round(canvas_size_pixel).astype(np.int32))


        return self.explorer.render_bitmap_mask(patch_box_meter,patch_angle,
                                                layer_names=layer_names,canvas_size_pixel=canvas_size_pixel,
                                                figsize=figsize,n_row=n_row)

    def render_bitmap_rgb(self,
                            patch_box_meter:Tuple[float,float,float,float],
                            patch_angle:float,
                            layer_names:List[str] = None,
                            figsize:Tuple[int,int] = (15,15),
                            n_row:int = 2) -> Tuple[Figure,List[Axes]]:
        """
        Render RGB-PNG图像 of the patch specified by patch_box_meter and patch_angle.
        :param Patch patch_box_meter_meter :定义显示图框(补丁框)大小 [x_center,y_center,height,width],UTM坐标系,单位:米.
        :param patch_angle: 显示图框(补丁框)的方向角,单位:deg,正北为0 deg.
        :param layer_names:['intersection','road']或者 None
        :param canvas_size_pixel:Size of the outputs mask (h,w).
        :param figsize:Size of the figure.
        :param n_row:Number of rows with plots. 绘图的行数
        :return:The matplotlib figure and a list of axes of the rendered layers.
        """
        # 计算 patch_box_meter对应像素的宽*高,单位pix
        canvas_size_pixel = np.array((patch_box_meter[2],patch_box_meter[3])) * self.scale_PixelPerMeter
        canvas_size_pixel = tuple(np.round(canvas_size_pixel).astype(np.int32))


        return self.explorer.render_bitmap_rgb(patch_box_meter,patch_angle,
                                                layer_names=layer_names,canvas_size_pixel=canvas_size_pixel,
                                                figsize=figsize,n_row=n_row)

    def extract_polygon(self,polygon_token:str) -> Polygon:
        """
        Construct a shapely Polygon object out of a polygon token.用多边形标记构造一个形状良好的多边形对象.
        :param polygon_token:The token of the polygon record.
        :return:The polygon wrapped in a shapely Polygon object.
        """
        return self.explorer.extract_polygon(polygon_token)  # 得到对应的多边形

    
    def get_polygon_token_using_node(self,x:float,y:float,polygons:Dict=None) -> str:
        """依据(x,y)定位该点属于哪个多边形内部

        Args:
            x (float):_description_
            y (float):_description_
            polygons (Dict,optional):_description_. Defaults to None.

        Returns:
            _type_:多边形 token
        """
        # 逐个检查多边形
        if polygons is None:
            polygons =self.polygon
        for polygon in polygons:
            link_node_tokens = polygon["link_node_tokens"]
            # 将多边形的节点坐标提取为NumPy数组
            polygon_nodes = np.array([(node["x"],node["y"]) for node_token in link_node_tokens
                                    for node in self.node if node["token"] == node_token])
            # 使用了射线法来确定点是否在多边形内部.该实现假设多边形的边界是封闭的,即首尾节点连接形成闭合的多边形 
            if self._is_point_in_polygon(x,y,polygon_nodes):
                return polygon["token"]
            
        # 如果点不在任何一个多边形内部,引发一个异常
        raise Exception(f"({x},{y})不属于任何 polygon,请检查.")
    
    
    def get_dubinspose_token_from_polygon(self,veh_pose:Tuple[float,float,float],polygon_token:str) -> str:
        """根据veh_pose(x,y,yaw)车辆定位位姿,从polygon_token所属的dubinspose list中匹配lane_connector的起始dubinspose.

        前提:
        1. vehicle所处的polygon为进入交叉路口前的一个road多边形.
        2. veh_pose在polygon内部.

        匹配原则:
        - Step 1:筛除掉|航向夹角|>0.5*Pi.
        - Step 2:
        - Case1:对于剩下2个dubinspose情况:
            - Case1-1:longitudinal_error 全部>=0,选择|longitudinal_error|大的.
            - Case1-2:longitudinal_error一正一负,选择longitudinal_error正的.
            - Case1-3:longitudinal_error 全部<=0,选择|longitudinal_error|小的.
        - Case2:对于剩下1个dubinspose情况——选择该dubinspose.

        Args:
            veh_pose (Tuple[float,float,float]):车辆的位姿.
            polygon_token (str):指定的polygon token.

        Returns:
            str:最佳匹配的dubinspose_token.
        """
        if self.location == "jiangxi_jiangtong":
            if polygon_token == 'polygon-18':
                return "'dubinspose-282'"

        if not polygon_token.startswith('polygon-'):
            raise ValueError(f"Invalid polygon_token:{polygon_token}")

        id_polygon = int(polygon_token.split('-')[1])
        if id_polygon > len(self.polygon):
            raise IndexError(f"Polygon ID {id_polygon} out of bounds.请检查.")

        link_dubinspose_tokens = self.polygon[id_polygon]['link_dubinspose_tokens']
        if not link_dubinspose_tokens:
            raise ValueError('No dubinspose tokens available for the specified polygon.')

        dubinsposes_indicators = []
        for token in link_dubinspose_tokens:
            id_dubinspose = int(token.split('-')[1])
            lateral_error,longitudinal_error,heading_error_rad = utils.compute_two_pose_error(
                self.dubins_pose[id_dubinspose]['x'],
                self.dubins_pose[id_dubinspose]['y'],
                self.dubins_pose[id_dubinspose]['yaw'],
                *veh_pose) # 每个dubinspose到veh_pose的关系
            
            distance = math.sqrt(lateral_error**2 + longitudinal_error**2)
            indicators = {
                "token_dubinspose":token,
                "lateral_error":lateral_error,
                "longitudinal_error":longitudinal_error,
                "heading_error_rad":heading_error_rad,
                "distance":distance
            }# 四个位姿关系指标
            dubinsposes_indicators.append(indicators)

        filtered_dubinsposes = [i for i,indicator in enumerate(dubinsposes_indicators)
                            if abs(indicator['heading_error_rad']) <= 0.5 * math.pi]

        if len(filtered_dubinsposes) == 1:
            return dubinsposes_indicators[filtered_dubinsposes[0]]["token_dubinspose"]
        elif len(filtered_dubinsposes) == 2:
            longitudinal_errors = dict()
            for index in filtered_dubinsposes:
                        longitudinal_errors[index] =  dubinsposes_indicators[index]['longitudinal_error']
            # 根据三种情况选择最佳的dubinspose
            if all(value >= 0 for value in longitudinal_errors.values()):
                 # case1-1:longitudinal_error 全部>=0,选择 |longitudinal_error|大的
                best_index = max(longitudinal_errors,key=lambda k:abs(longitudinal_errors[k]))

            elif all(value <= 0 for value in longitudinal_errors.values()):
                # case1-3:longitudinal_error 全部<=0,选择 |longitudinal_error|小的key
                best_index = min(longitudinal_errors,key=lambda k:abs(longitudinal_errors[k]))
            else:
                # case1-2:longitudinal_error 一正一负,选择 longitudinal_error 正的
                best_index = next(index for index,value in longitudinal_errors.items() if value > 0)

            return dubinsposes_indicators[best_index]["token_dubinspose"]

        raise ValueError(f"Unexpected condition with filtered node poses. Please check.")
        
        
        
    def get_dubinspose_token_from_polygon_old(self,veh_pose:Tuple[float,float,float],polygon_token:str)-> str:
        """功能:依据veh_pose(x,y,yaw)车辆定位位姿,从polygon_token所属的dubinspose list中匹配lane_connector的起始 dubinspose
        前提:1)vehicle 所处的polygon为进入交叉路口前的一个road 多边形.
                2)veh_pose在polygon内部.
        原理:匹配原则 
                -step 1:筛除掉 |航向夹角|>0.5*Pi;
                -step 2:-case1:对于剩下2个dubinspose情况:
                            -case1-1:longitudinal_error 全部>=0,选择 |longitudinal_error|大的;(无全部为0情况)
                            -case1-2:longitudinal_error 一正一负,选择 longitudinal_error 正的;
                            -case1-3:longitudinal_error 全部=<0,选择 |longitudinal_error|大的;(无全部为0情况)
                        -case2:对于剩下1个dubinspose情况——选择该dubinspose
        """       
        if polygon_token.startswith('polygon-'):
            id_polygon = int(polygon_token.split('-')[1])
            if id_polygon <=len(self.polygon):
                link_dubinspose_tokens = self.polygon[id_polygon]['link_dubinspose_tokens']
                if len(link_dubinspose_tokens) == 0:
                    raise Exception('No dubinspose tokens available for the specified polygon.')
                
                dubinspose_token_best_matched = None
                dubinsposes_indicators = [{} for _ in range(len(link_dubinspose_tokens))]
                for index,value in enumerate(link_dubinspose_tokens):
                    id_dubinspose = int(value.split('-')[1])
                    lateral_error,longitudinal_error,heading_error_rad=utils.compute_two_pose_error(
                                                    self.dubins_pose[id_dubinspose]['x'],
                                                    self.dubins_pose[id_dubinspose]['y'],
                                                    self.dubins_pose[id_dubinspose]['yaw'],
                                                    veh_pose[0],veh_pose[1],veh_pose[2]) #每个dubinspose到veh_pose的关系
                    distance = math.sqrt(lateral_error ** 2 + longitudinal_error ** 2)
                    dubinsposes_indicators[index]= {"token_dubinspose":value,
                                                "lateral_error":lateral_error,
                                                "longitudinal_error":longitudinal_error,
                                                "heading_error_rad":heading_error_rad,
                                                "distance":distance 
                                                }# 四个位姿关系指标
                
                filtered_dubinsposes = [ i for i,indicator in enumerate(dubinsposes_indicators)
                                    if abs(indicator['heading_error_rad']) <= 0.5 * math.pi]
                
                if len(filtered_dubinsposes)==1:
                    dubinspose_token_best_matched = dubinsposes_indicators[filtered_dubinsposes[0]]["token_dubinspose"]    
                
                elif len(filtered_dubinsposes)==2:
                    # 获取所有的longitudinal_error
                    longitudinal_errors = dict()
                    # longitudinal_errors = [dubinsposes_indicators[index]['longitudinal_error'] for index in filtered_dubinsposes]
                    for index in filtered_dubinsposes:
                        longitudinal_errors[index] =  dubinsposes_indicators[index]['longitudinal_error']
                        
                    
                    # 判断经过筛选后的longitudinal_error是否都是正的或都是负的
                    all_positive = all(err >= 0 for err in longitudinal_errors.values())
                    all_negative = all(err <= 0 for err in longitudinal_errors.values())

                    # 根据三种情况选择最佳的dubinspose
                    if all_positive:
                        # case1-1:longitudinal_error 全部>=0,选择 |longitudinal_error|大的
                        best_index = max(longitudinal_errors,key=longitudinal_errors.get)
                    elif all_negative:
                        # case1-3:longitudinal_error 全部<=0,选择 |longitudinal_error|大的key
                        best_index = min(longitudinal_errors,key=longitudinal_errors.get)
                    else:
                        # case1-2:longitudinal_error 一正一负,选择 longitudinal_error 正的
                        best_index = [index for index,err in longitudinal_errors.items() if err > 0][0]
                    dubinspose_token_best_matched = dubinsposes_indicators[best_index]["token_dubinspose"]    
                    
                else:
                    raise Exception(f"dubinspose_token_best_matched ={dubinspose_token_best_matched}有问题,请检查.") 
                if dubinspose_token_best_matched == None:
                    raise Exception(f"dubinspose_token_best_matched ={dubinspose_token_best_matched}有问题,请检查.")
                return dubinspose_token_best_matched
            
            else:
                raise Exception(f"polygon_token ={polygon_token},id_polygon ={id_polygon}超出界限,请检查.")
        else:
            raise Exception(f"polygon_token ={polygon_token}有问题,请检查.")
        
    @staticmethod
    def _is_point_in_polygon(x,y,polygon_nodes):
        # 将点坐标组合为一个坐标数组
        point = np.array([x,y])
        # 创建一个线段列表,每个线段连接多边形的两个相邻节点
        segments = [(polygon_nodes[i],polygon_nodes[(i + 1) % len(polygon_nodes)]) for i in range(len(polygon_nodes))]
        # 射线法判断点是否在多边形内部
        inside = False
        for segment in segments:
            # 提取线段的两个端点
            p1,p2 = segment
            if (p1[1] <= y < p2[1]) or (p2[1] <= y < p1[1]):
                # 计算线段与射线的交点的x坐标
                x_intersect = (p2[0] - p1[0]) * (y - p1[1]) / (p2[1] - p1[1]) + p1[0]
                if x < x_intersect:
                    inside = not inside
        return inside
    
    @staticmethod
    def _calculate_distance_of_two_nodes(node1:Tuple[float,float],node2:Tuple[float,float]) -> float:
        
        x1,y1 = node1
        x2,y2 = node2
        distance = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        return distance
        
    
    def _get_start_dubinspose_of_all_intersection(self,):
        pass
    
    # TODO
    def get_outgoing_plogon_token(self,):
        pass
    
    # TODO
    def get_outgoing_dubinspose_token(self,):
        pass
    
    # TODO
    def get_outgoing_dubinspose_token(self,):
        pass
    
    # TODO 增加新的脚本,处理所有的road中lane的连接,并将所有的path修改为lane;1) lane connectors ,交叉路口连通性;2)lane;
    # TODO 注意,nuscenes中 有关于 离散化采样 lane centerlines的代码 nuscenes-devkit/python-sdk/nuscenes/map_expansion/arcline_path_utils.py
    
    # TODO 
    def render_centerlines_all(self,
                            resolution_meters:float = 0.5,
                            figsize:Union[None,float,Tuple[float,float]] = None,
                            bitmap:Optional[BitMap] = None) -> Tuple[Figure,Axes]:
        """
        Render the centerlines of all lanes and lane connectors.
        渲染 所有的 path 和 connectivity_pathsample 的中心线centerlines;
        TODO 后续再实现渲染部分token确定的 lane centerlines
        """
        # 渲染 all lanes
        # self.explorer.render_centerlines(resolution_meters=resolution_meters,figsize=figsize,bitmap=bitmap)
        # 渲染 all lane connectors
        self.explorer.render_connector_path_centerlines(resolution_meters=resolution_meters,figsize=figsize,bitmap=bitmap)
    
    def _extract_connector_path(self)-> Tuple[list,list]:
        """ 
        """
        temp_reference_path = list()
        temp_reference_path = self.connectivity_path.copy()
        temp_intersection = list()
        temp_intersection = self.intersection.copy()
        temp_dubins_pose = list()
        temp_dubins_pose = self.dubins_pose.copy()
                
        for pathsample_key,pathsample_value in self.connector_path.items():
            # 1)处理 temp_reference_path
            pathsample_number = int( pathsample_key.split('-')[1]) 
            path_token = 'path-' + str(pathsample_number)  # 构建对应的path的token
            if 'pathsample' not in temp_reference_path[pathsample_number].keys():
                # 将pathsample的value加入到connectivity_path中对应的pathsample_number
                temp_reference_path[pathsample_number]['pathsample'] =dict()
                temp_reference_path[pathsample_number]['pathsample']['link_insection_token']=pathsample_value['link_insection_token']
                temp_reference_path[pathsample_number]['pathsample']['dubins_curve']=pathsample_value['dubins_curve']
                temp_reference_path[pathsample_number]['pathsample']['path_points']=np.array(pathsample_value['path_points'])
            
            #  2) 处理 temp_intersection
            intersection_number = int( pathsample_value['link_insection_token'].split('-')[1])# 'intersection-0'提取0
            if 'link_connectivity_path_tokens' not in temp_intersection[intersection_number]:
                temp_intersection[intersection_number]['link_connectivity_path_tokens'] = list() 
            temp_intersection[intersection_number]['link_connectivity_path_tokens'].append(path_token) 
            
            # 3) 处理 temp_dubins_pose
            start_dubinspose_token = pathsample_value['dubins_curve'][0]['start_dubinspose_token']
            start_dubinspose_number = int( start_dubinspose_token.split('-')[1])
            if self.dubins_pose[start_dubinspose_number]['is_first_dubinsnode']  is not True:
                raise Exception(f"dubins_pose 图层,token= {start_dubinspose_token} 有问题,请检查.")
            if 'link_connectivity_path_tokens' not in temp_dubins_pose[start_dubinspose_number]:
                temp_dubins_pose[start_dubinspose_number]['link_connectivity_path_tokens'] = list() 
            temp_dubins_pose[start_dubinspose_number]['link_connectivity_path_tokens'].append(path_token)
            
        return temp_reference_path,temp_intersection,temp_dubins_pose
        
        
        
        

class TgScenesMapExplorer:
    """ Helper class to explore the tgScenes map data. """
    def __init__(self,
                 map_api:TgScenesMap,
                 representative_layers:Tuple[str] = ('intersection','road','unloading_area','loading_area'),
                 color_map:dict = None):
        """
        :param map_api:TgScenesMap database class.
        :param representative_layers:代表整个地图数据的图层.
        :param color_map:Color map.
        """
        # Mutable default argument.
        if color_map is None:
            color_map = dict(road='#b2df8a',
                             road_block='#2A2A2A',
                             intersection='#fb9a99',
                             loading_area='#6a3d9a',
                             unloading_area='#7e772e',
                             connector_path='#1f78b4',
                             base_path='#1f78b4',
                             dubins_pose='#064b7a')
        
        self.map_api = map_api
        self.representative_layers = representative_layers
        self.color_map = color_map

        self.canvas_max_x = self.map_api.semantic_map_json['bitmap_rgb_PNG']['canvas_edge_meter'][0]# 画布的地理尺寸
        self.canvas_min_x = 0
        self.canvas_max_y = self.map_api.semantic_map_json['bitmap_rgb_PNG']['canvas_edge_meter'][1]
        self.canvas_min_y = 0
        self.canvas_aspect_ratio = (self.canvas_max_x - self.canvas_min_x) / (self.canvas_max_y - self.canvas_min_y)#画布横纵比


    def render_bitmap_mask(self,
                        patch_box_meter:Tuple[float,float,float,float],
                        patch_angle:float,
                        layer_names:List[str],
                        canvas_size_pixel:Tuple[int,int],
                        figsize:Tuple[int,int],
                        n_row:int = 2) -> Tuple[Figure,List[Axes]]:
        """
        渲染栅格图中的 bitmap_rgb.png
        the patch specified by patch_box_meter and patch_angle.
        :param Patch box :定义显示图框(补丁框)大小 [x_center,y_center,height,width],单位:米.
        :param patch_angle: 显示图框(补丁框)的方向角,单位:deg,正北为0 deg.
        :param layer_names:A list of layer names to be extracted.
        :param canvas_size_pixel:Size of the outputs mask (width,height). 画布的像素尺寸,宽*高,单位 pix .
        :param figsize:Size of the figure.
        :param n_row:Number of rows with plots.
        :return:The matplotlib figure and a list of axes of the rendered layers.
        """
        if layer_names is None:
            layer_names = self.map_api.non_geometric_layers

        map_mask = self.get_bitmap_mask(patch_box_meter,patch_angle,layer_names,canvas_size)

        # If no canvas_size is specified,retrieve the default from the outputs of get_map_mask.
        if canvas_size is None:
            canvas_size = map_mask.shape[1:]

        fig = plt.figure(figsize=figsize)
        ax = fig.add_axes([0,0,1,1])
        ax.set_xlim(0,canvas_size[1])
        ax.set_ylim(0,canvas_size[0])

        n_col = len(map_mask) // n_row
        gs = gridspec.GridSpec(n_row,n_col)
        gs.update(wspace=0.025,hspace=0.05)
        for i in range(len(map_mask)):
            r = i // n_col
            c = i - r * n_col
            subax = plt.subplot(gs[r,c])
            subax.imshow(map_mask[i],origin='lower')
            subax.text(canvas_size[0] * 0.5,canvas_size[1] * 1.1,layer_names[i])
            subax.grid(False)

        return fig,fig.axes
    
     
    def render_layers(self,
                      layer_names:List[str],
                      alpha:float,
                      figsize:Union[None,float,Tuple[float,float]],
                      tokens:List[str] = None,
                      bitmap:Optional[BitMap] = None) -> Tuple[Figure,Axes]:
        """
        Render a list of layers.
        :param layer_names:A list of layer names.
        :param alpha:The opacity of each layer.
        :param figsize:Size of the whole figure.
        :param tokens:Optional list of tokens to render. None means all tokens are rendered.
        :param bitmap:Optional BitMap object to render below the other map layers.
        :return:The matplotlib figure and axes of the rendered layers.
        """
        figsize1 =self._get_figsize(figsize)
        fig = plt.figure(figsize=figsize1)
        # fig = plt.figure(self._get_figsize(figsize))
        ax = fig.add_axes([0,0,1,1 / self.canvas_aspect_ratio])

        ax.set_xlim(self.canvas_min_x,self.canvas_max_x)
        ax.set_ylim(self.canvas_min_y,self.canvas_max_y)

        if bitmap is not None:
            if bitmap.bitmap_type == 'bitmap_mask':
                bitmap.render_mask_map(ax)
            elif bitmap.bitmap_type == 'bitmap_rgb':
                bitmap.render_rgb_map(ax)
            else:
                raise Exception('###Exception### 非法的 bitmap type:%s' % self.bitmap_type) # 自定义异常

        layer_names = list(set(layer_names))
        for layer_name in layer_names:
            self._render_layer(ax,layer_name,alpha,tokens)

        ax.legend()

        return fig,ax

    def render_map_patch(self,
                         box_coords:Tuple[float,float,float,float],
                         layer_names:List[str] = None,
                         alpha:float = 0.5,
                         figsize:Tuple[float,float] = (15,15),
                         render_egoposes_range:bool = True,
                         render_legend:bool = True,
                         bitmap:Optional[BitMap] = None,
                         fig:plt.figure = None) -> Tuple[Figure,Axes]:
        """
        Renders a rectangular patch specified by `box_coords`. By default renders all layers.
        渲染一个矩形图框,指定矩形框的xy坐标范围.
        :param box_coords:The rectangular patch coordinates (x_min,y_min,x_max,y_max). 
                矩形框的局部 x ,y 坐标范围
        :param layer_names:All the non geometric layers that we want to render.
        :param alpha:The opacity of each layer.
                每一个图层的透明度.
        :param figsize:Size of the whole figure.
                整张fig的尺寸
        :param render_egoposes_range:是否渲染一个包围所有自身姿态( ego poses)的矩形.
        :param render_legend:Whether to render the legend of map layers.
        :param bitmap:Optional BitMap object to render below the other map layers.
        :return:The matplotlib figure and axes of the rendered layers.
        """
        x_min,y_min,x_max,y_max = box_coords

        # if layer_names is None:
        #     layer_names = self.map_api.non_geometric_layers

        if fig is None:
            fig = plt.figure(figsize=figsize)

        local_width = x_max - x_min
        local_height = y_max - y_min
        assert local_height > 0,'Error:Map patch has 0 height!'
        local_aspect_ratio = local_width / local_height #局部图框的横纵比例

        ax = fig.add_axes([0,0,1,1 / local_aspect_ratio]) #创建一个子图(axes)对象,并指定其位置和大小

        if bitmap is not None:
            if bitmap.bitmap_type == 'bitmap_mask':
                bitmap.render_mask_map(ax)
            elif self.bitmap_type == 'bitmap_rgb':
                bitmap.render_rgb_map(ax)
            else:
                raise Exception('###Exception### 非法的 bitmap type:%s' % self.bitmap_type) # 自定义异常
            
        for layer_name in layer_names:#渲染各图层
            self._render_layer(ax,layer_name,alpha)

        x_margin = np.minimum(local_width / 4,50) #xy追加边距
        y_margin = np.minimum(local_height / 4,10)
        ax.set_xlim(x_min - x_margin,x_max + x_margin)
        ax.set_ylim(y_min - y_margin,y_max + y_margin)

        if render_egoposes_range:
            ax.add_patch(Rectangle((x_min,y_min),local_width,local_height,fill=False,linestyle='-.',color='red',
                                   lw=3))
            ax.text(x_min + local_width / 100,y_min + local_height / 2,"%g m" % local_height,
                    color='red',fontsize=14,weight='bold')
            ax.text(x_min + local_width / 2,y_min + local_height / 100,"%g m" % local_width,
                    color='red',fontsize=14,weight='bold') #fig上添加文字标注

        if render_legend:
            ax.legend(frameon=True,loc='upper right')

        return fig,ax

    def extract_polygon(self,polygon_token:str) -> Polygon:
        """
        Construct a shapely Polygon object out of a polygon token. 用多边形标记构造一个形状良好的多边形对象.
        :param polygon_token:The token of the polygon record.
        :return:The polygon wrapped in a shapely Polygon object.
        """
        polygon_record = self.map_api.get('polygon',polygon_token)  # 得到了对应的多边形记录

        link_coords = [(self.map_api.get('node',token)['x'],self.map_api.get('node',token)['y'])
                           for token in polygon_record['link_node_tokens']]  # 多边形坐标列表

        return Polygon(link_coords)  # 返回多边形
    
    @staticmethod
    def get_patch_coord(patch_box_meter:Tuple[float,float,float,float],
                        patch_angle:float = 0.0) -> Polygon:
        """
        Convert patch_box_meter to shapely Polygon coordinates.
        :param Patch box :定义显示图框(补丁框)大小 [x_center,y_center,height,width],单位:米.
        :param patch_angle: 显示图框(补丁框)的方向角,单位:deg,正北为0 deg.
        :return:Box Polygon for patch_box_meter.
        """
        patch_x,patch_y,patch_h,patch_w = patch_box_meter

        x_min = patch_x - patch_w / 2.0
        y_min = patch_y - patch_h / 2.0
        x_max = patch_x + patch_w / 2.0
        y_max = patch_y + patch_h / 2.0

        patch = box(x_min,y_min,x_max,y_max)
        patch = affinity.rotate(patch,patch_angle,origin=(patch_x,patch_y),use_radians=False)

        return patch

    def _get_figsize(self,figsize:Union[None,float,Tuple[float,float]]) -> Tuple[float,float]:
        """
        Utility function that scales the figure size by the map canvas size.按地图画布大小缩放图形大小的实用函数.
        If figsize is:
        - None      => Return default scale.
        - Scalar    => Scale canvas size.
        - Two-tuple => Use the specified figure size.
        :param figsize:The input figure size.
        :return:The outputs figure size.
        """
        # 将画布大小除以任意标量得到厘米范围.
        # canvas_size_pixel = np.array(self.map_api.canvas_edge_meter)[::-1] / 200
        canvas_size_pixel = np.array(self.map_api.canvas_edge_meter) / 200

        if figsize is None:
            return tuple(canvas_size_pixel)
        elif type(figsize) in [int,float]:
            return tuple(canvas_size_pixel * figsize)
        elif type(figsize) == tuple and len(figsize) == 2:
            return figsize
        else:
            raise Exception('Error:Invalid figsize:%s' % figsize)


    def _render_layer(self,ax:Axes,layer_name:str,alpha:float,tokens:List[str] = None) -> None:
        """
        Wrapper method that renders individual layers on an axis.
        :param ax:The matplotlib axes where the layer will get rendered.
        :param layer_name:Name of the layer that we are interested in.
        :param alpha:The opacity of the layer to be rendered.
        :param tokens:Optional list of tokens to render. None means all tokens are rendered.
        """
        if layer_name in self.map_api.non_geometric_polygon_layers:
            self._render_polygon_layer(ax,layer_name,alpha,tokens)  # road、intersection、loading_area以及unloading_area绘制完毕
        elif layer_name in self.map_api.non_geometric_line_layers:
            # self._render_line_layer(ax,layer_name,alpha,tokens)#!待调试
            pass
        elif layer_name == "road_block":
            alpha = 1
            self._render_road_block_polygon_layer(ax,layer_name,alpha,tokens)  # road_block绘制完毕
        else:
            raise ValueError("{} is not a valid layer".format(layer_name))


    def _render_polygon_layer(self,ax:Axes,layer_name:str,alpha:float,tokens:List[str] = None) -> None:
        """
        Renders an individual non-geometric polygon layer on an axis.
        :param ax:The matplotlib axes where the layer will get rendered.
        :param layer_name:Name of the layer that we are interested in.
        :param alpha:The opacity of the layer to be rendered.
        :param tokens:Optional list of tokens to render. None means all tokens are rendered.
        """ 
        if layer_name not in self.map_api.non_geometric_polygon_layers:
            raise ValueError('{} is not a polygonal layer'.format(layer_name))

        first_time = True
        records = getattr(self.map_api,layer_name) # 获得 某个图层'road'的内容,list
        if tokens is not None:
            records = [r for r in records if r['token'] in tokens]
        else:
            for record in records:  # 一个token一个的画图
                polygon = self.map_api.extract_polygon(record['link_polygon_token'])  # 创建一个 Shapely 多边形对象

                if first_time:
                    label = layer_name
                    first_time = False
                else:
                    label = None
                base_color = to_rgba(self.color_map[layer_name]) # Convert the hex to RGBA
                edge_color_with_alpha = (base_color[0],base_color[1],base_color[2],1)  # Set edge alpha to 1
                face_color_with_alpha = (base_color[0],base_color[1],base_color[2],0.1)  # Set face alpha to 0.1
                
                ax.add_patch(descartes.PolygonPatch(polygon,fc=face_color_with_alpha,ec=edge_color_with_alpha,label=label))
                # ax.add_patch(descartes.PolygonPatch(polygon,fc=self.color_map[layer_name],alpha=alpha,label=label) #!之前的代码
                
                
    def _render_road_block_polygon_layer(self,ax:Axes,layer_name:str,alpha:float,tokens:List[str] = None) -> None:
        """ 渲染road_block图层.是一个多边形,表示road is blocked
        """ 
        if layer_name != 'road_block':
            raise ValueError('{} is error,must road_block'.format(layer_name))

        first_time = True
        records = getattr(self.map_api,layer_name) # 获得 某个图层'road'的内容,list
        if tokens is not None:
            records = [r for r in records if r['token'] in tokens]
        else:
            for record in records:#一个token一个的画图
                link_nodeblock_tokens=record['link_nodeblock_tokens']
                link_coords = [(self.map_api.get('node_block',token)['x'],self.map_api.get('node_block',token)['y'])
                           for token in link_nodeblock_tokens] #组成 road_block 多边形的角点
                polygon =  Polygon(link_coords)# 创建一个 Shapely 多边形对象

                if first_time:
                    label = layer_name
                    first_time = False
                else:
                    label = None
                base_color = to_rgba(self.color_map[layer_name]) # Convert the hex to RGBA
                edge_color_with_alpha = (base_color[0],base_color[1],base_color[2],1)  # Set edge alpha to 1
                face_color_with_alpha = (base_color[0],base_color[1],base_color[2],0.5)  # Set face alpha to 0.1
                
                ax.add_patch(descartes.PolygonPatch(polygon,fc=face_color_with_alpha,ec=edge_color_with_alpha,label=label))            
   
                
    def render_base_path_centerlines(self,
                                    ax:Axes,
                                    alpha:float=0.5,
                                    resampling_rate:float = 0.2) -> None:
        """绘制参考路径,base_path.
        1)按照离散间隔绘制waypoints点;2)绘制所有dubinspose 
        
        Args:
            ax (Axes): x轴
            alpha (float, optional): 透明度. Defaults to 0.8.
            resampling_rate (float, optional): waypoints离散间隔采样. Defaults to 0.1.
                注:在这里,step = int(1 / resampling_rate)计算了降采样步长.
                如果resampling_rate是0.5,那么step值为2,所以我们每隔一个点取一个点来绘图.
                如果resampling_rate是0.25,那么step值为4,意味着我们每隔3个点取一个点.
                如果resampling_rate是0.2,那么step值为5,意味着我们每隔4个点取一个点.
                如果resampling_rate是0.1,那么step值为10,意味着我们每隔9个点取一个点.
        """
        for id_path,reference_path in enumerate(self.map_api.reference_path): 
            if reference_path['type'] != 'base_path':
                continue 
                    
            self.render_dubins_poses(ax=ax,alpha=0.4,dubinspose_tokens=reference_path['link_dubinspose_tokens'])
            if "waypoints" not in reference_path:   
                continue         
            waypoints = reference_path['waypoints']
            # Resampling
            step = int( 1 / resampling_rate )
            waypoints_resampling = np.array( waypoints[::step] )
            temp_color = self.color_map['connector_path']
            if len(waypoints_resampling) > 0:
                ax.plot(waypoints_resampling[:,0],waypoints_resampling[:,1],color=temp_color,linestyle='--',alpha=alpha,lw=1.2)
                
                    
    def render_connector_path_centerlines(self,
                                           ax:Axes,
                                           alpha:float=0.8,
                                           resampling_rate:float = 0.1) -> None:
        """绘制参考路径,connector_path.
        1)按照离散间隔绘制waypoints点;2)绘制所有dubinspose 
        
        Args:
            ax (Axes): x轴
            alpha (float, optional): 透明度. Defaults to 0.8.
            resampling_rate (float, optional): waypoints离散间隔采样. Defaults to 0.1.
                注:在这里,step = int(1 / resampling_rate)计算了降采样步长.
                如果resampling_rate是0.5,那么step值为2,所以我们每隔一个点取一个点来绘图.
                如果resampling_rate是0.25,那么step值为4,意味着我们每隔3个点取一个点.
                如果resampling_rate是0.2,那么step值为5,意味着我们每隔4个点取一个点.
                如果resampling_rate是0.1,那么step值为10,意味着我们每隔9个点取一个点.
        """
        
        for id_path,reference_path in enumerate(self.map_api.reference_path): 
            if reference_path['type'] != 'connector_path':
                continue
                         
            self.render_dubins_poses(ax=ax,alpha=0.4,dubinspose_tokens=reference_path['link_dubinspose_tokens'])  # 绘制dubinspose点及箭头
            if "waypoints" not in reference_path:   
                continue         
            waypoints = reference_path['waypoints']
            # Resampling
            step = int( 1 / resampling_rate )
            waypoints_resampling = np.array( waypoints[::step] )
            temp_color = self.color_map['connector_path']
            if len(waypoints_resampling) > 0:
                ax.plot(waypoints_resampling[:,0],waypoints_resampling[:,1],color=temp_color,linestyle='-.',alpha=alpha,lw=0.5)
   

            
    def render_dubins_poses(self,ax:Axes,alpha:float=0.5,dubinspose_tokens:List[str] = None) -> None:
        """绘制所有dubinspose.

        Args:
            ax (Axes): x轴
            alpha (float, optional): 透明度. Defaults to 0.5.
            dubinspose_tokens (List[str], optional): 杜宾斯点序列. Defaults to None.
        """
        for token_dubinspose in dubinspose_tokens:
            id_dubinspose = int( token_dubinspose.split('-')[1]) 
            if id_dubinspose > len(self.map_api.dubins_pose):
                continue # 语句被用来告诉 Python 跳过当前循环块中的剩余语句,然后继续进行下一轮循环.          
            x = self.map_api.dubins_pose[id_dubinspose]['x']
            y = self.map_api.dubins_pose[id_dubinspose]['y']
            yaw = self.map_api.dubins_pose[id_dubinspose]['yaw']
            temp_color = self.color_map['dubins_pose']
            # 绘制箭头dubinspose
            node_ux = x + 3*np.cos(yaw)
            node_vy = y + 3*np.sin(yaw)
            ax.arrow(x,y,node_ux-x,node_vy-y,color=temp_color,alpha=alpha,head_width=0.65,head_length=1.0)




if __name__ == "__main__":
    # 测试代码
    pass
     







 